/**
 * @author Bernhard Arnold
 *         Takashi Matsushita
 * @date:  2015-03-12
 */

#ifndef tmEventSetup_esTriggerMenu_hh
#define tmEventSetup_esTriggerMenu_hh

#include "tmEventSetup/esScale.hh"
#include "tmEventSetup/esCondition.hh"
#include "tmEventSetup/esAlgorithm.hh"

#include <map>
#include <string>

namespace tmeventsetup
{

/**
 *  This class implements data structure for TriggerMenu
 */
class esTriggerMenu
{
  public:
    esTriggerMenu()
      : algorithm_map_(), condition_map_(), scale_map_(),
#if defined(SWIG)
        algorithm_map_p_(), condition_map_p_(), scale_map_p_(),
#endif
        external_map_(), token_to_condition_(), name_(), version_(), comment_(),
        datetime_(), uuid_firmware_(), scale_set_name_(), n_modules_(), version(0) { };

    virtual ~esTriggerMenu() = default;

    /** get algorithm_map_
     *
     * @return reference of algorithm_map_
     */
    const std::map<std::string, esAlgorithm>& getAlgorithmMap() const { return algorithm_map_; };

    /** get condition_map_
     *
     * @return reference of condition_map_
     */
    const std::map<std::string, esCondition>& getConditionMap() const { return condition_map_; };

    /** get scale_map_
     *
     * @return reference of scale_map_
     */
    const std::map<std::string, esScale>& getScaleMap() const { return scale_map_; };

#if defined(SWIG)
    const std::map<std::string, tmeventsetup::esAlgorithm*>& getAlgorithmMapPtr() const { return algorithm_map_p_; };
    const std::map<std::string, tmeventsetup::esCondition*>& getConditionMapPtr() const { return condition_map_p_; };
    const std::map<std::string, tmeventsetup::esScale*>& getScaleMapPtr() const { return scale_map_p_; };
#endif

    /** get menu name
     *
     * @return name of the menu
     */
    const std::string& getName() const { return name_; };

    /** get grammar version
     *
     * @return grammar version
     */
    const std::string& getVersion() const { return version_; };

    /** get comment on the menu
     *
     * @return comment
     */
    const std::string& getComment() const { return comment_; };

    /** get datetime of the menu
     *
     * @return datetime
     */
    const std::string& getDatetime() const { return datetime_; };

#if defined(SWIG)
    /** get UUID of menu generated by trigger menu editor
     *
     * @return UUID
     */
    const std::string& getMenuUuid() const { return uuid_menu_; };
#endif

    /** get UUID of firmware generated by VHDL producer
     *
     * @return UUID
     */
    const std::string& getFirmwareUuid() const { return uuid_firmware_; };

    /** get scale set name
     *
     * @return scale set name
     */
    const std::string& getScaleSetName() const { return scale_set_name_; };

    /** get number of uGT boards for the menu
     *
     * @return number of uGT boards
     */
    const unsigned int getNmodules() const { return n_modules_; };

    /** set menu name
     *
     * @param value [in] name of the menu
     */
    void setName(const std::string& value) { name_ = value; };

    /** set grammar version
     *
     * @param value [in] grammar version
     */
    void setVersion(const std::string& value) { version_ = value; };

    /** set comment on the menu
     *
     * @param value [in] comment
     */
    void setComment(const std::string& value) { comment_ = value; };

    /** set datetime of the menu
     *
     * @param value [in] datetime
     */
    void setDatetime(const std::string& value) { datetime_ = value; };

#if defined(SWIG)
    /** set UUID of firmware generated by VHDL producer
     *
     * @param value [in] UUID
     */
    void setMenuUuid(const std::string& value) { uuid_menu_ = value; };
#endif

    /** set UUID of firmware generated by VHDL producer
     *
     * @param value [in] UUID
     */
    void setFirmwareUuid(const std::string& value) { uuid_firmware_ = value; };

    /** set scale set name
     *
     * @param value [in] scale set name
     */
    void setScaleSetName(const std::string& value) { scale_set_name_ = value; };

    /** set number of uGT boards for the menu
     *
     * @param value [in] number of uGT boards
     */
    void setNmodules(const unsigned int value) { n_modules_ = value; };

  protected:
    std::map<std::string, esAlgorithm> algorithm_map_;      /**< map of algorithm <algorithm name, esAlgorithm> */
    std::map<std::string, esCondition> condition_map_;      /**< map of condition <condition name, esCondition> */
    std::map<std::string, esScale> scale_map_;              /**< map of scale <scale name, esScale*> */
#if defined(SWIG)
    std::map<std::string, esAlgorithm*> algorithm_map_p_;
    std::map<std::string, esCondition*> condition_map_p_;
    std::map<std::string, esScale*> scale_map_p_;
#endif
    std::map<std::string, unsigned int> external_map_;      /**< map of external <external name, channel id> */
    std::map<std::string, std::string> token_to_condition_; /**< look-up table for translating expression in grammar to expression in condition */
    std::string name_;                                      /**< name of the menu */
    std::string version_;                                   /**< menu grammar version */
    std::string comment_;                                   /**< comment on the menu */
    std::string datetime_;                                  /**< datetime of the menu */
#if defined(SWIG)
    std::string uuid_menu_;                                 /**< uuid of menu */
#endif
    std::string uuid_firmware_;                             /**< uuid of firmware */
    std::string scale_set_name_;                            /**< scale set name */
    unsigned int n_modules_;                                /**< number of uGT boards for the menu */
    unsigned int version;
};

} // namespace tmeventsetup

#endif // tmEventSetup_esTriggerMenu_hh
